@@ -22,6 +22,7 @@ import numpy as np
 import aesara
 from aesara.configdefaults import config
 from aesara.graph.basic import Constant, Variable
+from aesara.link.utils import get_destroy_dependencies
 
 
 logger = logging.getLogger("aesara.compile.profiling")
@@ -216,7 +217,7 @@ class ProfileStats:
     #
 
     vm_call_time = 0.0
-    # Total time spent in Function.fn.__call__
+    # Total time spent in Function.vm.__call__
     #
 
     apply_time = None
@@ -780,7 +781,7 @@ class ProfileStats:
         )
         if self.fct_call_time > 0:
             print(
-                f"  Time in Function.fn.__call__: {self.vm_call_time}s ({100 * self.vm_call_time / self.fct_call_time:.3f}%)",
+                f"  Time in Function.vm.__call__: {self.vm_call_time}s ({100 * self.vm_call_time / self.fct_call_time:.3f}%)",
                 file=file,
             )
             local_time = sum(self.apply_time.values())
@@ -1035,12 +1036,14 @@ class ProfileStats:
                     if isinstance(val, Constant):
                         compute_map[val][0] = 1
 
+            destroy_dependencies = get_destroy_dependencies(fgraph)
+
             # Initial executable_nodes
             executable_nodes = set()
             for var in fgraph.inputs:
                 for c, _ in fgraph.clients[var]:
                     if c != "output":
-                        deps = c.inputs + c.destroy_dependencies
+                        deps = c.inputs + destroy_dependencies[c]
                         if all(compute_map[v][0] for v in deps):
                             executable_nodes.add(c)
 
@@ -1163,10 +1166,12 @@ class ProfileStats:
                         # smaller, stop this iteration, move to next node
                         done_dict[frozen_set] = max_mem_count
 
+                        destroy_dependencies = get_destroy_dependencies(fgraph)
+
                         for var in node.outputs:
                             for c, _ in fgraph.clients[var]:
                                 if c != "output":
-                                    deps = c.inputs + c.destroy_dependencies
+                                    deps = c.inputs + destroy_dependencies[c]
                                     if all(compute_map[v][0] for v in deps):
                                         new_exec_nodes.add(c)
 
