@@ -1,15 +1,19 @@
 """Core graph classes."""
+import abc
 import warnings
 from collections import deque
 from copy import copy
 from itertools import count
 from typing import (
     TYPE_CHECKING,
+    Any,
     Callable,
     Collection,
     Deque,
     Dict,
     Generator,
+    Generic,
+    Hashable,
     Iterable,
     Iterator,
     List,
@@ -39,9 +43,15 @@ from aesara.misc.ordered_set import OrderedSet
 
 
 if TYPE_CHECKING:
+    from aesara.graph.op import Op
     from aesara.graph.type import Type
 
 
+OpType = TypeVar("OpType", bound="Op")
+OptionalApplyType = TypeVar("OptionalApplyType", None, "Apply", covariant=True)
+_TypeType = TypeVar("_TypeType", bound="Type")
+_IdType = TypeVar("_IdType", bound=Hashable)
+
 T = TypeVar("T", bound="Node")
 NoParams = object()
 NodeAndChildren = Tuple[T, Optional[Iterable[T]]]
@@ -55,7 +65,6 @@ class Node(MetaObject):
     keeps track of its parents via `Variable.owner` / `Apply.inputs`.
 
     """
-    type: "Type"
     name: Optional[str]
 
     def get_parents(self):
@@ -68,7 +77,7 @@ class Node(MetaObject):
         raise NotImplementedError()
 
 
-class Apply(Node):
+class Apply(Node, Generic[OpType]):
     """A `Node` representing the application of an operation to inputs.
 
     Basically, an `Apply` instance is an object that represents the
@@ -94,34 +103,31 @@ class Apply(Node):
 
     Attributes
     ----------
-    op : Op
+    op
         The operation that produces `outputs` given `inputs`.
-    inputs : List[Variable]
+    inputs
         The arguments of the expression modeled by the `Apply` node.
-    outputs : List[Variable]
+    outputs
         The outputs of the expression modeled by the `Apply` node.
 
     """
 
-    def __init__(self, op, inputs, outputs):
-        """
-        Parameters
-        ----------
-        op : Op
-        inputs : List[Variable]
-        outputs : List[Variable]
+    def __init__(
+        self,
+        op: OpType,
+        inputs: Sequence["Variable"],
+        outputs: Sequence["Variable"],
+    ):
+        if not isinstance(inputs, Sequence):
+            raise TypeError("The inputs of an Apply must be a sequence type")
+
+        if not isinstance(outputs, Sequence):
+            raise TypeError("The output of an Apply must be a sequence type")
 
-        """
         self.op = op
         self.inputs: List[Variable] = []
         self.tag = Scratchpad()
 
-        if not isinstance(inputs, (list, tuple)):
-            raise TypeError("The inputs of an Apply must be a list or tuple")
-
-        if not isinstance(outputs, (list, tuple)):
-            raise TypeError("The output of an Apply must be a list or tuple")
-
         # filter inputs to make sure each element is a Variable
         for input in inputs:
             if isinstance(input, Variable):
@@ -200,28 +206,40 @@ class Apply(Node):
     def __repr__(self):
         return str(self)
 
-    def clone(self):
-        """
-        Duplicate this Apply instance with inputs = self.inputs.
+    def clone(self, clone_inner_graph: bool = False) -> "Apply[OpType]":
+        r"""Clone this `Apply` instance.
+
+        Parameters
+        ----------
+        clone_inner_graph
+            If ``True``, clone `HasInnerGraph` `Op`\s and their inner-graphs.
 
         Returns
         -------
-        object
-            A new Apply instance (or subclass instance) with new outputs.
+        A new `Apply` instance  with new outputs.
 
         Notes
         -----
-        Tags are copied from self to the returned instance.
+        Tags are copied from `self` to the returned instance.
 
         """
+        from aesara.graph.op import HasInnerGraph
+
+        new_op = self.op
+
+        if isinstance(new_op, HasInnerGraph) and clone_inner_graph:  # type: ignore
+            new_op = new_op.clone()  # type: ignore
+
         cp = self.__class__(
-            self.op, self.inputs, [output.clone() for output in self.outputs]
+            new_op, self.inputs, [output.clone() for output in self.outputs]
         )
         cp.tag = copy(self.tag)
         return cp
 
-    def clone_with_new_inputs(self, inputs, strict=True):
-        """Duplicate this `Apply` instance in a new graph.
+    def clone_with_new_inputs(
+        self, inputs: Sequence["Variable"], strict=True, clone_inner_graph=False
+    ) -> "Apply[OpType]":
+        r"""Duplicate this `Apply` instance in a new graph.
 
         Parameters
         ----------
@@ -236,6 +254,8 @@ class Apply(Node):
             ``self.outputs``.  If ``False``, then there's no guarantee that the
             clone's outputs will have the same types as ``self.outputs``,
             and cloning may not even be possible (it depends on the `Op`).
+        clone_inner_graph : bool
+            If ``True``, clone `HasInnerGraph` `Op`\s and their inner-graphs.
 
         Returns
         -------
@@ -243,9 +263,11 @@ class Apply(Node):
             An `Apply` instance with the same `Op` but different outputs.
 
         """
+        from aesara.graph.op import HasInnerGraph
+
         assert isinstance(inputs, (list, tuple))
         remake_node = False
-        new_inputs = inputs[:]
+        new_inputs: List["Variable"] = list(inputs)
         for i, (curr, new) in enumerate(zip(self.inputs, new_inputs)):
             if curr.type != new.type:
                 if strict:
@@ -258,10 +280,15 @@ class Apply(Node):
                     remake_node = True
 
         if remake_node:
-            new_node = self.op.make_node(*new_inputs)
+            new_op = self.op
+
+            if isinstance(new_op, HasInnerGraph) and clone_inner_graph:  # type: ignore
+                new_op = new_op.clone()  # type: ignore
+
+            new_node = new_op.make_node(*new_inputs)
             new_node.tag = copy(self.tag).__update__(new_node.tag)
         else:
-            new_node = self.clone()
+            new_node = self.clone(clone_inner_graph=clone_inner_graph)
             new_node.inputs = new_inputs
         return new_node
 
@@ -288,7 +315,7 @@ class Apply(Node):
         return self.op.params_type
 
 
-class Variable(Node):
+class Variable(Node, Generic[_TypeType, OptionalApplyType]):
     r"""
     A :term:`Variable` is a node in an expression graph that represents a
     variable.
@@ -386,42 +413,51 @@ class Variable(Node):
     # __slots__ = ['type', 'owner', 'index', 'name']
     __count__ = count(0)
 
-    _owner: Optional[Apply]
+    _owner: OptionalApplyType
 
     @property
-    def owner(self) -> Optional[Apply]:
+    def owner(self) -> OptionalApplyType:
         return self._owner
 
     @owner.setter
     def owner(self, value) -> None:
         self._owner = value
 
+    @property
+    def index(self):
+        return self._index
+
+    @index.setter
+    def index(self, value):
+        self._index = value
+
     def __init__(
         self,
-        type,
-        owner: Optional[Apply] = None,
+        type: _TypeType,
+        owner: OptionalApplyType,
         index: Optional[int] = None,
         name: Optional[str] = None,
-    ):
+    ) -> None:
         super().__init__()
 
         self.tag = ValidatingScratchpad("test_value", type.filter)
 
         self.type = type
 
-        if owner is not None and not isinstance(owner, Apply):
-            raise TypeError("owner must be an Apply instance", owner)
         self._owner = owner
 
+        if owner is not None and not isinstance(owner, Apply):
+            raise TypeError("owner must be an Apply instance")
+
         if index is not None and not isinstance(index, int):
-            raise TypeError("index must be an int", index)
+            raise TypeError("index must be an int")
         self.index = index
 
         if name is not None and not isinstance(name, str):
-            raise TypeError("name must be a string", name)
+            raise TypeError("name must be a string")
         self.name = name
 
-        self.auto_name = "auto_" + str(next(self.__count__))
+        self.auto_name = f"auto_{next(self.__count__)}"
 
     def get_test_value(self):
         """Get the test value.
@@ -475,22 +511,18 @@ class Variable(Node):
         return "\n".join(to_print)
 
     def clone(self):
-        """Return a new `Variable` like `self`.
+        """Return a new, un-owned `Variable` like `self`.
 
         Returns
         -------
         Variable instance
-            A new `Variable` instance (or subclass instance) with no owner or
-            index.
+            A new `Variable` instance  with no owner or index.
 
         Notes
         -----
-        Tags are copied to the returned instance.
-
-        Name is copied to the returned instance.
+        Tags and names are copied to the returned instance.
 
         """
-        # return copy(self)
         cp = self.__class__(self.type, None, None, self.name)
         cp.tag = copy(self.tag)
         return cp
@@ -586,7 +618,98 @@ class Variable(Node):
         return d
 
 
-class Constant(Variable):
+class AtomicVariable(Variable[_TypeType, None]):
+    """A node type that has no ancestors and should never be considered an input to a graph."""
+
+    def __init__(self, type: _TypeType, **kwargs):
+        super().__init__(type, None, None, **kwargs)
+
+    @abc.abstractmethod
+    def signature(self):
+        ...
+
+    def merge_signature(self):
+        return self.signature()
+
+    def equals(self, other):
+        """
+        This does what `__eq__` would normally do, but `Variable` and `Apply`
+        should always be hashable by `id`.
+        """
+        return isinstance(other, type(self)) and self.signature() == other.signature()
+
+    @property
+    def owner(self):
+        return None
+
+    @owner.setter
+    def owner(self, value):
+        if value is not None:
+            raise ValueError("AtomicVariable instances cannot have an owner.")
+
+    @property
+    def index(self):
+        return None
+
+    @index.setter
+    def index(self, value):
+        if value is not None:
+            raise ValueError("AtomicVariable instances cannot have an index.")
+
+
+class NominalVariable(AtomicVariable[_TypeType]):
+    """A variable that enables alpha-equivalent comparisons."""
+
+    __instances__: Dict[Tuple["Type", Hashable], "NominalVariable"] = {}
+
+    def __new__(cls, id: _IdType, typ: _TypeType, **kwargs):
+        if (typ, id) not in cls.__instances__:
+            var_type = typ.variable_type
+            type_name = f"Nominal{var_type.__name__}"
+
+            def _reduce(self):
+                return cls, (self.id, self.type)
+
+            def _str(self):
+                return f"*{self.id}-{var_type.__str__(self)}"
+
+            new_type = type(
+                type_name, (cls, var_type), {"__reduce__": _reduce, "__str__": _str}
+            )
+            res: NominalVariable = super().__new__(new_type)
+
+            cls.__instances__[(typ, id)] = res
+
+        return cls.__instances__[(typ, id)]
+
+    def __init__(self, id: _IdType, typ: _TypeType, **kwargs):
+        self.id = id
+        super().__init__(typ, **kwargs)
+
+    def clone(self):
+        return self
+
+    def __eq__(self, other):
+        if self is other:
+            return True
+
+        return (
+            type(self) == type(other)
+            and self.id == other.id
+            and self.type == other.type
+        )
+
+    def __hash__(self):
+        return hash((type(self), self.id, self.type))
+
+    def __repr__(self):
+        return f"{type(self).__name__}({repr(self.id)}, {repr(self.type)})"
+
+    def signature(self) -> Tuple[_TypeType, _IdType]:
+        return (self.type, self.id)
+
+
+class Constant(AtomicVariable[_TypeType]):
     """A `Variable` with a fixed `data` field.
 
     `Constant` nodes make numerous optimizations possible (e.g. constant
@@ -601,24 +724,17 @@ class Constant(Variable):
 
     # __slots__ = ['data']
 
-    def __init__(self, type, data, name=None):
-        super().__init__(type, None, None, name)
+    def __init__(self, type: _TypeType, data: Any, name: Optional[str] = None):
+        super().__init__(type, name=name)
         self.data = type.filter(data)
         add_tag_trace(self)
 
     def get_test_value(self):
         return self.data
 
-    def equals(self, other):
-        # this does what __eq__ should do, but Variable and Apply should always be hashable by id
-        return isinstance(other, Constant) and self.signature() == other.signature()
-
     def signature(self):
         return (self.type, self.data)
 
-    def merge_signature(self):
-        return self.signature()
-
     def __str__(self):
         if self.name is not None:
             return self.name
@@ -641,9 +757,9 @@ class Constant(Variable):
         if value is not None:
             raise ValueError("Constant instances cannot have an owner.")
 
-    value = property(lambda self: self.data, doc="read-only data access method")
-
-    # index is not defined, because the `owner` attribute must necessarily be None
+    @property
+    def value(self):
+        return self.data
 
 
 def walk(
@@ -847,6 +963,7 @@ def clone(
     outputs: List[Variable],
     copy_inputs: bool = True,
     copy_orphans: Optional[bool] = None,
+    clone_inner_graphs: bool = False,
 ) -> Tuple[Collection[Variable], Collection[Variable]]:
     r"""Copies the sub-graph contained between inputs and outputs.
 
@@ -862,6 +979,8 @@ def clone(
         When ``None``, use the `copy_inputs` value.
         When ``True``, new orphans nodes are created.
         When ``False``, original orphans nodes are reused in the new graph.
+    clone_inner_graphs : bool
+        If ``True``, clone `HasInnerGraph` `Op`\s and their inner-graphs.
 
     Returns
     -------
@@ -877,42 +996,109 @@ def clone(
     """
     if copy_orphans is None:
         copy_orphans = copy_inputs
-    equiv = clone_get_equiv(inputs, outputs, copy_inputs, copy_orphans)
+    equiv = clone_get_equiv(
+        inputs,
+        outputs,
+        copy_inputs=copy_inputs,
+        copy_orphans=copy_orphans,
+        clone_inner_graphs=clone_inner_graphs,
+    )
     return [cast(Variable, equiv[input]) for input in inputs], [
         cast(Variable, equiv[output]) for output in outputs
     ]
 
 
+def clone_node_and_cache(
+    node: Apply,
+    clone_d: Dict[Union[Apply, Variable, "Op"], Union[Apply, Variable, "Op"]],
+    clone_inner_graphs=False,
+    **kwargs,
+) -> Optional[Apply]:
+    """Clone an `Apply` node and cache the results in `clone_d`.
+
+    This function handles `Op` clones that are generated by inner-graph
+    cloning.
+
+    Returns
+    -------
+    ``None`` if all of `node`'s outputs are already in `clone_d`; otherwise,
+    return the clone of `node`.
+
+    """
+    if all(out in clone_d for out in node.outputs):
+        # If all of `node`'s outputs already have replacements or clones in
+        # `clone_d`, then there's likely no need to clone it
+        return None
+
+    # Use a cached `Op` clone when available
+    new_op: Optional["Op"] = cast(Optional["Op"], clone_d.get(node.op))
+
+    cloned_inputs: List[Variable] = [cast(Variable, clone_d[i]) for i in node.inputs]
+
+    new_node = node.clone_with_new_inputs(
+        cloned_inputs,
+        # Only clone inner-graph `Op`s when there isn't a cached clone (and
+        # when `clone_inner_graphs` is enabled)
+        clone_inner_graph=clone_inner_graphs if new_op is None else False,
+        **kwargs,
+    )
+
+    if new_op:
+        # If we didn't clone the inner-graph `Op` above, because
+        # there was a cached version, set the cloned `Apply` to use
+        # the cached clone `Op`
+        new_node.op = new_op
+
+    clone_d[node] = new_node
+
+    if new_node.op is not node.op:
+        clone_d.setdefault(node.op, new_node.op)
+
+    for old_o, new_o in zip(node.outputs, new_node.outputs):
+        clone_d.setdefault(old_o, new_o)
+
+    return new_node
+
+
 def clone_get_equiv(
     inputs: Sequence[Variable],
     outputs: Sequence[Variable],
     copy_inputs: bool = True,
     copy_orphans: bool = True,
-    memo: Optional[Dict[Node, Node]] = None,
-) -> Dict[Node, Node]:
-    """
-    Return a dictionary that maps from `Variable` and `Apply` nodes in the
-    original graph to a new node (a clone) in a new graph.
+    memo: Optional[
+        Dict[Union[Apply, Variable, "Op"], Union[Apply, Variable, "Op"]]
+    ] = None,
+    clone_inner_graphs: bool = False,
+) -> Dict[Union[Apply, Variable, "Op"], Union[Apply, Variable, "Op"]]:
+    r"""Clone the graph between `inputs` and `outputs` and return a map of the cloned objects.
 
-    This function works by recursively cloning inputs... rebuilding a directed
-    graph from the inputs up to eventually building new outputs.
+    This function works by recursively cloning inputs and rebuilding a directed
+    graph from the inputs up.
+
+    If `memo` already contains entries for some of the objects in the graph,
+    those objects are replaced with their values in `memo` and *not*
+    unnecessarily cloned.
 
     Parameters
     ----------
-    inputs : a list of Variables
-    outputs : a list of Variables
-    copy_inputs : bool
-        True means to create the cloned graph from new input
-        nodes (the bottom of a feed-upward graph).
-        False means to clone a graph that is rooted at the original input
-        nodes.
-    copy_orphans :
-        When ``True``, new constant nodes are created. When ``False``, original
-        constant nodes are reused in the new graph.
-    memo : None or dict
+    inputs
+        Inputs of the graph to be cloned.
+    outputs
+        Outputs of the graph to be cloned.
+    copy_inputs
+        ``True`` means to create the cloned graph from cloned input nodes.
+        ``False`` means to clone a graph that is rooted at the original input
+        nodes.  `Constant`\s are *not* cloned.
+    copy_orphans
+        When ``True``, inputs with no owners are cloned.  When ``False``,
+        original inputs are reused in the new graph.  Cloning is *not*
+        performed for `Constant`\s.
+    memo
         Optionally start with a partly-filled dictionary for the return value.
         If a dictionary is passed, this function will work in-place on that
         dictionary and return it.
+    clone_inner_graphs
+        If ``True``, clone `HasInnerGraph` `Op`\s and their inner-graphs.
 
     """
     if memo is None:
@@ -920,7 +1106,7 @@ def clone_get_equiv(
 
     # clone the inputs if necessary
     for input in inputs:
-        if copy_inputs:
+        if not isinstance(input, Constant) and copy_inputs:
             cpy = input.clone()
             cpy.owner = None
             cpy.index = None
@@ -932,16 +1118,13 @@ def clone_get_equiv(
     for apply in io_toposort(inputs, outputs):
         for input in apply.inputs:
             if input not in memo:
-                if copy_orphans:
+                if not isinstance(input, Constant) and copy_orphans:
                     cpy = input.clone()
                     memo[input] = cpy
                 else:
                     memo[input] = input
 
-        new_apply = apply.clone_with_new_inputs([memo[i] for i in apply.inputs])
-        memo.setdefault(apply, new_apply)
-        for output, new_output in zip(apply.outputs, new_apply.outputs):
-            memo.setdefault(output, new_output)
+        clone_node_and_cache(apply, memo, clone_inner_graphs=clone_inner_graphs)
 
     # finish up by cloning any remaining outputs (it can happen)
     for output in outputs:
@@ -952,12 +1135,11 @@ def clone_get_equiv(
 
 
 def clone_replace(
-    output: List[Variable],
+    output: Collection[Variable],
     replace: Optional[
         Union[Iterable[Tuple[Variable, Variable]], Dict[Variable, Variable]]
     ] = None,
-    strict: bool = True,
-    share_inputs: bool = True,
+    **rebuild_kwds,
 ) -> List[Variable]:
     """Clone a graph and replace subgraphs within it.
 
@@ -966,15 +1148,12 @@ def clone_replace(
 
     Parameters
     ----------
-    output : Aesara Variables (or Aesara expressions)
+    output
         Aesara expression that represents the computational graph.
-    replace : dict
+    replace
         Dictionary describing which subgraphs should be replaced by what.
-    share_inputs : bool
-        If ``True``, use the same inputs (and shared variables) as the original
-        graph. If ``False``, clone them. Note that cloned shared variables still
-        use the same underlying storage, so they will always have the same
-        value.
+    rebuild_kwds
+        Keywords to `rebuild_collect_shared`.
 
     """
     from aesara.compile.function.pfunc import rebuild_collect_shared
@@ -996,14 +1175,10 @@ def clone_replace(
         )
     tmp_replace = [(x, x.type()) for x, y in items]
     new_replace = [(x, y) for ((_, x), (_, y)) in zip(tmp_replace, items)]
-    _, _outs, _ = rebuild_collect_shared(
-        output, [], tmp_replace, [], strict, share_inputs
-    )
+    _, _outs, _ = rebuild_collect_shared(output, [], tmp_replace, [], **rebuild_kwds)
 
     # TODO Explain why we call it twice ?!
-    _, outs, _ = rebuild_collect_shared(
-        _outs, [], new_replace, [], strict, share_inputs
-    )
+    _, outs, _ = rebuild_collect_shared(_outs, [], new_replace, [], **rebuild_kwds)
 
     return cast(List[Variable], outs)
 
@@ -1379,13 +1554,12 @@ def view_roots(node: Variable) -> List[Variable]:
     owner = node.owner
     if owner is not None:
         try:
-            view_map = owner.op.view_map
-            view_map = {owner.outputs[o]: i for o, i in view_map.items()}
+            vars_to_views = {owner.outputs[o]: i for o, i in owner.op.view_map.items()}
         except AttributeError:
             return [node]
-        if node in view_map:
+        if node in vars_to_views:
             answer = []
-            for i in view_map[node]:
+            for i in vars_to_views[node]:
                 answer += view_roots(owner.inputs[i])
             return answer
         else:
@@ -1661,3 +1835,38 @@ def get_var_by_name(
             results += (var,)
 
     return results
+
+
+def replace_nominals_with_dummies(inputs, outputs):
+    """Replace nominal inputs with dummy variables.
+
+    When constructing a new graph with nominal inputs from an existing graph,
+    pre-existing nominal inputs need to be replaced with dummy variables
+    beforehand; otherwise, sequential ID ordering (i.e. when nominals are IDed
+    based on the ordered inputs to which they correspond) of the nominals could
+    be broken, and/or circular replacements could manifest.
+
+    FYI: This function assumes that all the nominal variables in the subgraphs
+    between `inputs` and `outputs` are present in `inputs`.
+
+    """
+    existing_nominal_replacements = {
+        i: i.type() for i in inputs if isinstance(i, NominalVariable)
+    }
+
+    if existing_nominal_replacements:
+        # Replace existing nominal variables, because we need to produce an
+        # inner-graph for which the nominal variable IDs correspond exactly
+        # to their input order
+        _ = clone_get_equiv(
+            inputs,
+            outputs,
+            copy_inputs=False,
+            copy_orphans=False,
+            memo=existing_nominal_replacements,
+        )
+
+        outputs = [existing_nominal_replacements[o] for o in outputs]
+        inputs = [existing_nominal_replacements[i] for i in inputs]
+
+    return inputs, outputs
