@@ -325,6 +325,17 @@ class Feature:
         """
         return OrderedDict()
 
+    def clone(self):
+        """Create a clone that can be attached to a new `FunctionGraph`.
+
+        This default implementation returns `self`, which carries the
+        assumption that the `Feature` is essentially stateless.  If a subclass
+        has state of its own that is in any way relative to a given
+        `FunctionGraph`, this method should be overridden with an
+        implementation that actually creates a fresh copy.
+        """
+        return self
+
 
 class Bookkeeper(Feature):
     def on_attach(self, fgraph):
@@ -389,6 +400,9 @@ class History(Feature):
         fgraph.checkpoint = GetCheckpoint(self, fgraph)
         fgraph.revert = partial(self.revert, fgraph)
 
+    def clone(self):
+        return type(self)()
+
     def unpickle(self, fgraph):
         fgraph.checkpoint = GetCheckpoint(self, fgraph)
         fgraph.revert = partial(self.revert, fgraph)
@@ -516,6 +530,9 @@ class ReplaceValidate(History, Validator):
         Validator.on_attach(self, fgraph)
         self.unpickle(fgraph)
 
+    def clone(self):
+        return type(self)()
+
     def unpickle(self, fgraph):
         History.unpickle(self, fgraph)
         Validator.unpickle(self, fgraph)
@@ -641,18 +658,19 @@ class NodeFinder(Bookkeeper):
         self.d = {}
 
     def on_attach(self, fgraph):
-        if self.fgraph is not None:
-            raise Exception(
-                "A NodeFinder instance can only serve one " "FunctionGraph."
-            )
         if hasattr(fgraph, "get_nodes"):
-            raise AlreadyThere(
-                "NodeFinder is already present or in conflict" " with another plugin."
-            )
+            raise AlreadyThere("NodeFinder is already present")
+
+        if self.fgraph is not None and self.fgraph != fgraph:
+            raise Exception("A NodeFinder instance can only serve one FunctionGraph.")
+
         self.fgraph = fgraph
         fgraph.get_nodes = partial(self.query, fgraph)
         Bookkeeper.on_attach(self, fgraph)
 
+    def clone(self):
+        return type(self)()
+
     def on_detach(self, fgraph):
         """
         Should remove any dynamically added functionality
@@ -746,15 +764,23 @@ class PreserveVariableAttributes(Feature):
 class NoOutputFromInplace(Feature):
     """Prevent `FunctionGraph` outputs within a range from being altered in-place."""
 
-    def __init__(self, first_output_idx=0, last_output_idx=None):
-        self.first_idx = first_output_idx
-        self.last_idx = last_output_idx
+    def __init__(self, protected_out_ids):
+        self.protected_out_ids = tuple(protected_out_ids)
+
+    def on_attach(self, fgraph):
+        if hasattr(fgraph, "_no_output_from_inplace"):
+            raise AlreadyThere(f"InnerGraphWatcher is already attached to {fgraph}.")
+
+        fgraph._no_output_from_inplace = self
+
+    def clone(self):
+        return type(self)(self.protected_out_ids)
 
     def validate(self, fgraph):
         if not hasattr(fgraph, "destroyers"):
             return True
 
-        for out in fgraph.outputs[self.first_idx : self.last_idx]:
+        for out in tuple(fgraph.outputs[i] for i in self.protected_out_ids):
 
             node = out.owner
 
@@ -772,3 +798,5 @@ class NoOutputFromInplace(Feature):
                     f"operations. This has prevented the output {out} from "
                     "being computed by modifying another variable in-place."
                 )
+
+        return True
