@@ -1,7 +1,7 @@
 import copy
 import sys
 import warnings
-from abc import abstractmethod
+from abc import ABC, abstractmethod
 from typing import (
     TYPE_CHECKING,
     Any,
@@ -12,6 +12,7 @@ from typing import (
     Sequence,
     Text,
     Tuple,
+    TypeVar,
     Union,
     cast,
 )
@@ -50,12 +51,14 @@ ThunkCallableType = Callable[
     [PerformMethodType, StorageMapType, ComputeMapType, Apply], None
 ]
 
+C = TypeVar("C", bound=Callable)
 
-class ThunkType(Protocol):
+
+class ThunkType(Protocol[C]):
     inputs: List[List[Optional[List[Any]]]]
     outputs: List[List[Optional[List[Any]]]]
     lazy: bool
-    __call__: ThunkCallableType
+    __call__: C
     perform: PerformMethodType
 
 
@@ -132,8 +135,7 @@ def compute_test_value(node: Apply):
     thunk.inputs = [storage_map[v] for v in node.inputs]
     thunk.outputs = [storage_map[v] for v in node.outputs]
 
-    required = thunk()
-    assert not required  # We provided all inputs
+    thunk()
 
     for output in node.outputs:
         # Check that the output has been computed
@@ -495,7 +497,7 @@ class Op(MetaObject):
         node: Apply,
         storage_map: StorageMapType,
         compute_map: ComputeMapType,
-        no_recycling: bool,
+        no_recycling: List[Variable],
         debug: bool = False,
     ) -> ThunkType:
         """Make a Python thunk.
@@ -506,7 +508,7 @@ class Op(MetaObject):
         node_input_storage = [storage_map[r] for r in node.inputs]
         node_output_storage = [storage_map[r] for r in node.outputs]
 
-        if debug:
+        if debug and hasattr(self, "debug_perform"):
             p = node.op.debug_perform
         else:
             p = node.op.perform
@@ -551,7 +553,7 @@ class Op(MetaObject):
         node: Apply,
         storage_map: StorageMapType,
         compute_map: ComputeMapType,
-        no_recycling: bool,
+        no_recycling: List[Variable],
         impl: Optional[Text] = None,
     ) -> ThunkType:
         r"""Create a thunk.
@@ -612,13 +614,16 @@ class _NoPythonOp(Op):
         raise NotImplementedError("No Python implementation is provided by this Op.")
 
 
-class HasInnerGraph:
+class HasInnerGraph(ABC):
     r"""A mixin for an `Op` that contain an inner graph."""
 
+    fgraph: "FunctionGraph"
+    """A `FunctionGraph` of the inner function."""
+
     @property
     @abstractmethod
     def fn(self) -> "Function":
-        """The inner function."""
+        """The compiled inner-graph function."""
 
     @property
     @abstractmethod
@@ -630,6 +635,10 @@ class HasInnerGraph:
     def inner_outputs(self) -> List[Variable]:
         """The inner function's outputs."""
 
+    @abstractmethod
+    def clone(self) -> Op:
+        """Clone the `Op` and its inner-graph."""
+
 
 def get_test_value(v: Any) -> Any:
     """Get the test value for `v`.
