@@ -4,7 +4,7 @@ import re
 import sys
 import traceback
 import warnings
-from collections import Counter
+from collections import Counter, defaultdict
 from keyword import iskeyword
 from operator import itemgetter
 from tempfile import NamedTemporaryFile
@@ -318,8 +318,6 @@ def raise_with_op(
         raise exc_value.with_traceback(exc_trace)
 
     trace = getattr(node.outputs[0].tag, "trace", ())
-    if not trace and hasattr(node.op, "tag"):
-        trace = getattr(node.op.tag, "trace", ())
 
     exc_value.__thunk_trace__ = trace
     exc_value.__op_instance__ = node
@@ -366,8 +364,6 @@ def raise_with_op(
             detailed_err_msg += "\nInputs type_num: %s" % str(
                 [getattr(getattr(i[0], "dtype", ""), "num", "") for i in thunk.inputs]
             )
-        if hasattr(node.op, "__input_name__"):
-            detailed_err_msg += f"\nInputs name: {node.op.__input_name__}\n"
 
         detailed_err_msg += f"\nOutputs clients: {clients}\n"
     else:
@@ -764,9 +760,9 @@ def fgraph_to_python(
 
         node_output_names = [unique_name(v) for v in node.outputs]
 
-        body_assigns.append(
-            f"# {node}\n{', '.join(node_output_names)} = {local_compiled_func_name}({', '.join(node_input_names)})"
-        )
+        assign_comment_str = f"{indent(str(node), '# ')}"
+        assign_str = f"{', '.join(node_output_names)} = {local_compiled_func_name}({', '.join(node_input_names)})"
+        body_assigns.append(f"{assign_comment_str}\n{assign_str}")
 
     fgraph_input_names = [unique_name(v) for v in fgraph.inputs]
     fgraph_output_names = [unique_name(v) for v in fgraph.outputs]
@@ -793,3 +789,22 @@ def {fgraph_name}({", ".join(fgraph_input_names)}):
     )
 
     return fgraph_def
+
+
+def get_destroy_dependencies(fgraph: FunctionGraph) -> Dict[Apply, List[Variable]]:
+    """Construct a ``dict`` of nodes to variables that are implicit dependencies induced by `Op.destroy_map` and `Op.view_map`
+
+    These variable dependencies are in contrast to each node's inputs, which
+    are _explicit_ dependencies.
+
+    The variables in the values of this ``dict`` would be impossible to compute
+    after the current key nodes are evaluated, because node.thunk() is going to
+    destroy a common input variable needed by whatever node owns each variable
+    in destroy_dependencies.
+    """
+    order = fgraph.orderings()
+    destroy_dependencies = defaultdict(lambda: [])
+    for node in fgraph.apply_nodes:
+        for prereq in order.get(node, []):
+            destroy_dependencies[node].extend(prereq.outputs)
+    return destroy_dependencies
