@@ -118,6 +118,7 @@ def debugprint(
     print_op_info: bool = False,
     print_destroy_map: bool = False,
     print_view_map: bool = False,
+    print_fgraph_inputs: bool = False,
 ) -> Union[str, IOBase]:
     r"""Print a computation graph as text to stdout or a file.
 
@@ -175,6 +176,8 @@ def debugprint(
         Whether to print the `destroy_map`\s of printed objects
     print_view_map
         Whether to print the `view_map`\s of printed objects
+    print_fgraph_inputs
+        Print the inputs of `FunctionGraph`\s.
 
     Returns
     -------
@@ -197,7 +200,8 @@ def debugprint(
     if used_ids is None:
         used_ids = dict()
 
-    results_to_print = []
+    inputs_to_print = []
+    outputs_to_print = []
     profile_list: List[Optional[Any]] = []
     order: List[Optional[List[Apply]]] = []  # Toposort
     smap: List[Optional[StorageMapType]] = []  # storage_map
@@ -209,26 +213,30 @@ def debugprint(
 
     for obj in lobj:
         if isinstance(obj, Variable):
-            results_to_print.append(obj)
+            outputs_to_print.append(obj)
             profile_list.append(None)
             smap.append(None)
             order.append(None)
         elif isinstance(obj, Apply):
-            results_to_print.extend(obj.outputs)
+            outputs_to_print.extend(obj.outputs)
             profile_list.extend([None for item in obj.outputs])
             smap.extend([None for item in obj.outputs])
             order.extend([None for item in obj.outputs])
         elif isinstance(obj, Function):
-            results_to_print.extend(obj.maker.fgraph.outputs)
+            if print_fgraph_inputs:
+                inputs_to_print.extend(obj.maker.fgraph.inputs)
+            outputs_to_print.extend(obj.maker.fgraph.outputs)
             profile_list.extend([obj.profile for item in obj.maker.fgraph.outputs])
             if print_storage:
-                smap.extend([obj.fn.storage_map for item in obj.maker.fgraph.outputs])
+                smap.extend([obj.vm.storage_map for item in obj.maker.fgraph.outputs])
             else:
                 smap.extend([None for item in obj.maker.fgraph.outputs])
             topo = obj.maker.fgraph.toposort()
             order.extend([topo for item in obj.maker.fgraph.outputs])
         elif isinstance(obj, FunctionGraph):
-            results_to_print.extend(obj.outputs)
+            if print_fgraph_inputs:
+                inputs_to_print.extend(obj.inputs)
+            outputs_to_print.extend(obj.outputs)
             profile_list.extend([getattr(obj, "profile", None) for item in obj.outputs])
             smap.extend([getattr(obj, "storage_map", None) for item in obj.outputs])
             topo = obj.toposort()
@@ -236,7 +244,7 @@ def debugprint(
         elif isinstance(obj, (int, float, np.ndarray)):
             print(obj, file=_file)
         elif isinstance(obj, (In, Out)):
-            results_to_print.append(obj.variable)
+            outputs_to_print.append(obj.variable)
             profile_list.append(None)
             smap.append(None)
             order.append(None)
@@ -268,10 +276,29 @@ N.B.:
 
     op_information = {}
 
-    for r, p, s, o in zip(results_to_print, profile_list, smap, order):
+    for r in inputs_to_print:
+        _debugprint(
+            r,
+            prefix="-",
+            depth=depth,
+            done=done,
+            print_type=print_type,
+            file=_file,
+            ids=ids,
+            inner_graph_ops=inner_graph_ops,
+            stop_on_name=stop_on_name,
+            used_ids=used_ids,
+            op_information=op_information,
+            parent_node=r.owner,
+            print_op_info=print_op_info,
+            print_destroy_map=print_destroy_map,
+            print_view_map=print_view_map,
+        )
+
+    for r, p, s, o in zip(outputs_to_print, profile_list, smap, order):
 
         if hasattr(r.owner, "op"):
-            if isinstance(r.owner.op, HasInnerGraph):
+            if isinstance(r.owner.op, HasInnerGraph) and r not in inner_graph_ops:
                 inner_graph_ops.append(r)
             if print_op_info:
                 op_information.update(op_debug_information(r.owner.op, r.owner))
@@ -352,13 +379,39 @@ N.B.:
                 print_view_map=print_view_map,
             )
 
-            for idx, i in enumerate(inner_outputs):
+            if print_fgraph_inputs:
+                for inp in inner_inputs:
+                    _debugprint(
+                        r=inp,
+                        prefix="-",
+                        depth=depth,
+                        done=done,
+                        print_type=print_type,
+                        file=_file,
+                        ids=ids,
+                        stop_on_name=stop_on_name,
+                        inner_graph_ops=inner_graph_ops,
+                        inner_to_outer_inputs=inner_to_outer_inputs,
+                        used_ids=used_ids,
+                        op_information=op_information,
+                        parent_node=s.owner,
+                        print_op_info=print_op_info,
+                        print_destroy_map=print_destroy_map,
+                        print_view_map=print_view_map,
+                        inner_graph_node=s.owner,
+                    )
+                inner_to_outer_inputs = None
 
-                if isinstance(getattr(i.owner, "op", None), HasInnerGraph):
-                    inner_graph_ops.append(i)
+            for out in inner_outputs:
+
+                if (
+                    isinstance(getattr(out.owner, "op", None), HasInnerGraph)
+                    and out not in inner_graph_ops
+                ):
+                    inner_graph_ops.append(out)
 
                 _debugprint(
-                    r=i,
+                    r=out,
                     prefix=new_prefix,
                     depth=depth,
                     done=done,
@@ -375,6 +428,7 @@ N.B.:
                     print_op_info=print_op_info,
                     print_destroy_map=print_destroy_map,
                     print_view_map=print_view_map,
+                    inner_graph_node=s.owner,
                 )
 
     if file is _file:
@@ -407,6 +461,7 @@ def _debugprint(
     op_information: Optional[Dict[Apply, Dict[Variable, str]]] = None,
     parent_node: Optional[Apply] = None,
     print_op_info: bool = False,
+    inner_graph_node: Optional[Apply] = None,
 ) -> IOBase:
     r"""Print the graph leading to `r`.
 
@@ -459,6 +514,8 @@ def _debugprint(
     print_op_info
         Print extra information provided by the relevant `Op`\s.  For example,
         print the tap information for `Scan` inputs and outputs.
+    inner_graph_node
+        The inner-graph node in which `r` is contained.
     """
     if depth == 0:
         return file
@@ -591,7 +648,10 @@ def _debugprint(
                     new_prefix_child = prefix_child + "  "
 
                 if hasattr(i, "owner") and hasattr(i.owner, "op"):
-                    if isinstance(i.owner.op, HasInnerGraph):
+                    if (
+                        isinstance(i.owner.op, HasInnerGraph)
+                        and i not in inner_graph_ops
+                    ):
                         inner_graph_ops.append(i)
 
                 _debugprint(
@@ -615,6 +675,7 @@ def _debugprint(
                     print_op_info=print_op_info,
                     print_destroy_map=print_destroy_map,
                     print_view_map=print_view_map,
+                    inner_graph_node=inner_graph_node,
                 )
     else:
 
@@ -644,18 +705,12 @@ def _debugprint(
 
             var_output = f"{var_output} -> {outer_id_str}"
 
-            # This is an inner-graph input, so we need to find the outer node
-            # it belongs to and get the extra information from that
-            for inner_graph in inner_graph_ops:
-                if outer_r in inner_graph.owner.inputs:
-                    node_info = op_information.get(inner_graph.owner)
-                    if node_info and r in node_info:
-                        var_output = f"{var_output} ({node_info[r]})"
-                        break
-
-        node_info = op_information.get(parent_node) or op_information.get(r.owner)
-        if node_info and r in node_info:
-            var_output = f"{var_output} ({node_info[r]})"
+        # TODO: This entire approach will only print `Op` info for two levels
+        # of nesting.
+        for node in dict.fromkeys([inner_graph_node, parent_node, r.owner]):
+            node_info = op_information.get(node)
+            if node_info and r in node_info:
+                var_output = f"{var_output} ({node_info[r]})"
 
         print(var_output, file=file)
 
@@ -1360,6 +1415,9 @@ def pydotprint(
     # it, we must copy it.
     outputs = list(outputs)
     if isinstance(fct, Function):
+
+        # TODO: Get rid of all this `expanded_inputs` nonsense and use
+        # `fgraph.update_mapping`
         function_inputs = zip(fct.maker.expanded_inputs, fgraph.inputs)
         for i, fg_ii in reversed(list(function_inputs)):
             if i.update is not None:
@@ -1537,7 +1595,7 @@ def pydotprint(
             if hasattr(scan_op.op, "_fn"):
                 to_print = scan_op.op.fn
             else:
-                to_print = scan_op.op.outputs
+                to_print = scan_op.op.inner_outputs
             pydotprint(
                 to_print,
                 new_name,
